@@ -1,3 +1,17 @@
+# Copyright 2014 Google Inc. All rights reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 """Manipulate access control lists that Cloud Storage provides.
 
 :class:`gcloud.storage.bucket.Bucket` has a getting method that creates
@@ -5,8 +19,8 @@ an ACL object under the hood, and you can interact with that using
 :func:`gcloud.storage.bucket.Bucket.acl`::
 
   >>> from gcloud import storage
-  >>> connection = storage.get_connection(project, email, key_path)
-  >>> bucket = connection.get_bucket(bucket_name)
+  >>> client = storage.Client()
+  >>> bucket = client.get_bucket(bucket_name)
   >>> acl = bucket.acl
 
 Adding and removing permissions can be done with the following methods
@@ -70,6 +84,13 @@ class _ACLEntity(object):
 
     This is a helper class that you likely won't ever construct
     outside of using the factor methods on the :class:`ACL` object.
+
+    :type entity_type: string
+    :param entity_type: The type of entity (ie, 'group' or 'user').
+
+    :type identifier: string
+    :param identifier: The ID or e-mail of the entity. For the special
+                       entity types (like 'allUsers') this is optional.
     """
 
     READER_ROLE = 'READER'
@@ -77,15 +98,6 @@ class _ACLEntity(object):
     OWNER_ROLE = 'OWNER'
 
     def __init__(self, entity_type, identifier=None):
-        """Entity constructor.
-
-        :type entity_type: string
-        :param entity_type: The type of entity (ie, 'group' or 'user').
-
-        :type identifier: string
-        :param identifier: The ID or e-mail of the entity. For the special
-                           entity types (like 'allUsers') this is optional.
-        """
         self.identifier = identifier
         self.roles = set([])
         self.type = entity_type
@@ -94,11 +106,11 @@ class _ACLEntity(object):
         if not self.identifier:
             return str(self.type)
         else:
-            return '{self.type}-{self.identifier}'.format(self=self)
+            return '{acl.type}-{acl.identifier}'.format(acl=self)
 
     def __repr__(self):
-        return '<ACL Entity: {self} ({roles})>'.format(
-            self=self, roles=', '.join(self.roles))
+        return '<ACL Entity: {acl} ({roles})>'.format(
+            acl=self, roles=', '.join(self.roles))
 
     def get_roles(self):
         """Get the list of roles permitted by this entity.
@@ -113,62 +125,69 @@ class _ACLEntity(object):
 
         :type role: string
         :param role: The role to add to the entity.
-
-        :rtype: :class:`_ACLEntity`
-        :returns: The entity class.
         """
         self.roles.add(role)
-        return self
 
     def revoke(self, role):
         """Remove a role from the entity.
 
         :type role: string
         :param role: The role to remove from the entity.
-
-        :rtype: :class:`_ACLEntity`
-        :returns: The entity class.
         """
         if role in self.roles:
             self.roles.remove(role)
-        return self
 
     def grant_read(self):
         """Grant read access to the current entity."""
-
-        return self.grant(_ACLEntity.READER_ROLE)
+        self.grant(_ACLEntity.READER_ROLE)
 
     def grant_write(self):
         """Grant write access to the current entity."""
-
-        return self.grant(_ACLEntity.WRITER_ROLE)
+        self.grant(_ACLEntity.WRITER_ROLE)
 
     def grant_owner(self):
         """Grant owner access to the current entity."""
-
-        return self.grant(_ACLEntity.OWNER_ROLE)
+        self.grant(_ACLEntity.OWNER_ROLE)
 
     def revoke_read(self):
         """Revoke read access from the current entity."""
-
-        return self.revoke(_ACLEntity.READER_ROLE)
+        self.revoke(_ACLEntity.READER_ROLE)
 
     def revoke_write(self):
         """Revoke write access from the current entity."""
-
-        return self.revoke(_ACLEntity.WRITER_ROLE)
+        self.revoke(_ACLEntity.WRITER_ROLE)
 
     def revoke_owner(self):
         """Revoke owner access from the current entity."""
-
-        return self.revoke(_ACLEntity.OWNER_ROLE)
+        self.revoke(_ACLEntity.OWNER_ROLE)
 
 
 class ACL(object):
     """Container class representing a list of access controls."""
 
+    _URL_PATH_ELEM = 'acl'
+    _PREDEFINED_QUERY_PARAM = 'predefinedAcl'
+
+    _PREDEFINED_ACLS = frozenset([
+        'private',
+        'project-private',
+        'public-read',
+        'public-read-write',
+        'authenticated-read',
+        'bucket-owner-read',
+        'bucket-owner-full-control',
+    ])
+    """See:
+    https://cloud.google.com/storage/docs/access-control#predefined-acl
+    """
+
     loaded = False
 
+    # Subclasses must override to provide these attributes (typically,
+    # as properties).
+    reload_path = None
+    save_path = None
+
     def __init__(self):
         self.entities = {}
 
@@ -220,7 +239,8 @@ class ACL(object):
         if not isinstance(entity, _ACLEntity):
             raise ValueError('Invalid dictionary: %s' % entity_dict)
 
-        return entity.grant(role)
+        entity.grant(role)
+        return entity
 
     def has_entity(self, entity):
         """Returns whether or not this ACL has any entries for an entity.
@@ -228,7 +248,7 @@ class ACL(object):
         :type entity: :class:`_ACLEntity`
         :param entity: The entity to check for existence in this ACL.
 
-        :rtype: bool
+        :rtype: boolean
         :returns: True of the entity exists in the ACL.
         """
         self._ensure_loaded()
@@ -344,85 +364,85 @@ class ACL(object):
         self._ensure_loaded()
         return list(self.entities.values())
 
-    def reload(self):
+    @property
+    def client(self):
+        """Abstract getter for the object client."""
+        raise NotImplementedError
+
+    def _require_client(self, client):
+        """Check client or verify over-ride.
+
+        :type client: :class:`gcloud.storage.client.Client` or ``NoneType``
+        :param client: the client to use.  If not passed, falls back to the
+                       ``client`` stored on the current ACL.
+
+        :rtype: :class:`gcloud.storage.client.Client`
+        :returns: The client passed in or the currently bound client.
+        """
+        if client is None:
+            client = self.client
+        return client
+
+    def reload(self, client=None):
         """Reload the ACL data from Cloud Storage.
 
-        :rtype: :class:`ACL`
-        :returns: The current ACL.
+        :type client: :class:`gcloud.storage.client.Client` or ``NoneType``
+        :param client: Optional. The client to use.  If not passed, falls back
+                       to the ``client`` stored on the ACL's parent.
         """
-        raise NotImplementedError
+        path = self.reload_path
+        client = self._require_client(client)
 
-    def save(self, acl=None):
-        """A method to be overridden by subclasses.
-
-        :type acl: :class:`gcloud.storage.acl.ACL`, or a compatible list.
-        :param acl: The ACL object to save.  If left blank, this will save
-                    current entries.
-
-        :raises: NotImplementedError
-        """
-        raise NotImplementedError
-
-    def clear(self):
-        """Remove all entities from the ACL."""
-        raise NotImplementedError
-
-
-class BucketACL(ACL):
-    """An ACL specifically for a bucket."""
-
-    _URL_PATH_ELEM = 'acl'
-
-    def __init__(self, bucket):
-        """
-        :type bucket: :class:`gcloud.storage.bucket.Bucket`
-        :param bucket: The bucket to which this ACL relates.
-        """
-        super(BucketACL, self).__init__()
-        self.bucket = bucket
-
-    def reload(self):
-        """Reload the ACL data from Cloud Storage.
-
-        :rtype: :class:`gcloud.storage.acl.BucketACL`
-        :returns: The current ACL.
-        """
         self.entities.clear()
 
-        url_path = '%s/%s' % (self.bucket.path, self._URL_PATH_ELEM)
-        found = self.bucket.connection.api_request(method='GET', path=url_path)
+        found = client.connection.api_request(method='GET', path=path)
         self.loaded = True
-        for entry in found['items']:
+        for entry in found.get('items', ()):
             self.add_entity(self.entity_from_dict(entry))
 
-        return self
-
-    def save(self, acl=None):
-        """Save this ACL for the current bucket.
-
-        If called without arguments, this will save the entries
-        currently stored on this ACL::
-
-           >>> acl.save()
-
-        You can also provide a specific ACL to save instead of the one
-        currently set on the Bucket object::
-
-           >>> acl.save(acl=my_other_acl)
-
-        You can use this to set access controls to be consistent from
-        one bucket to another::
-
-          >>> bucket1 = connection.get_bucket(bucket1_name)
-          >>> bucket2 = connection.get_bucket(bucket2_name)
-          >>> bucket2.acl.save(bucket1.acl)
+    def _save(self, acl, predefined, client):
+        """Helper for :meth:`save` and :meth:`save_predefined`.
 
         :type acl: :class:`gcloud.storage.acl.ACL`, or a compatible list.
         :param acl: The ACL object to save.  If left blank, this will save
                     current entries.
 
-        :rtype: :class:`gcloud.storage.acl.BucketACL`
-        :returns: The current ACL.
+        :type predefined: string or None
+        :param predefined: An identifier for a predefined ACL.  Must be one
+                           of the keys in :attr:`_PREDEFINED_ACLS`
+                           If passed, `acl` must be None.
+
+        :type client: :class:`gcloud.storage.client.Client` or ``NoneType``
+        :param client: Optional. The client to use.  If not passed, falls back
+                       to the ``client`` stored on the ACL's parent.
+        """
+        query_params = {'projection': 'full'}
+        if predefined is not None:
+            acl = []
+            query_params[self._PREDEFINED_QUERY_PARAM] = predefined
+
+        path = self.save_path
+        client = self._require_client(client)
+        result = client.connection.api_request(
+            method='PATCH',
+            path=path,
+            data={self._URL_PATH_ELEM: list(acl)},
+            query_params=query_params)
+        self.entities.clear()
+        for entry in result.get(self._URL_PATH_ELEM, ()):
+            self.add_entity(self.entity_from_dict(entry))
+        self.loaded = True
+
+    def save(self, acl=None, client=None):
+        """Save this ACL for the current bucket.
+
+        :type acl: :class:`gcloud.storage.acl.ACL`, or a compatible list.
+        :param acl: The ACL object to save.  If left blank, this will save
+                    current entries.
+
+        :type client: :class:`gcloud.storage.client.Client` or ``NoneType``
+        :param client: Optional. The client to use.  If not passed, falls back
+                       to the ``client`` stored on the ACL's parent.
         """
         if acl is None:
             acl = self
@@ -431,18 +451,26 @@ class BucketACL(ACL):
             save_to_backend = True
 
         if save_to_backend:
-            result = self.bucket.connection.api_request(
-                method='PATCH', path=self.bucket.path,
-                data={self._URL_PATH_ELEM: list(acl)},
-                query_params={'projection': 'full'})
-            self.entities.clear()
-            for entry in result[self._URL_PATH_ELEM]:
-                self.add_entity(self.entity_from_dict(entry))
-            self.loaded = True
+            self._save(acl, None, client)
 
-        return self
+    def save_predefined(self, predefined, client=None):
+        """Save this ACL for the current bucket using a predefined ACL.
 
-    def clear(self):
+        :type predefined: string
+        :param predefined: An identifier for a predefined ACL.  Must be one
+                           of the keys in :attr:`_PREDEFINED_ACLS`
+                           If passed, `acl` must be None.
+
+        :type client: :class:`gcloud.storage.client.Client` or ``NoneType``
+        :param client: Optional. The client to use.  If not passed, falls back
+                       to the ``client`` stored on the ACL's parent.
+        """
+        if predefined not in self._PREDEFINED_ACLS:
+            raise ValueError("Invalid predefined ACL: %s" % (predefined,))
+
+        self._save(None, predefined, client)
+
+    def clear(self, client=None):
         """Remove all ACL entries.
 
         Note that this won't actually remove *ALL* the rules, but it
@@ -450,89 +478,69 @@ class BucketACL(ACL):
         have access to a bucket that you created even after you clear
         ACL rules with this method.
 
-        For example, imagine that you granted access to this bucket to a
-        bunch of coworkers::
-
-          >>> acl.user('coworker1@example.org').grant_read()
-          >>> acl.user('coworker2@example.org').grant_read()
-          >>> acl.save()
-
-        Now they work in another part of the company and you want to
-        'start fresh' on who has access::
-
-          >>> acl.clear()
-
-        At this point all the custom rules you created have been removed.
-
-        :rtype: :class:`gcloud.storage.acl.BucketACL`
-        :returns: The current ACL.
+        :type client: :class:`gcloud.storage.client.Client` or ``NoneType``
+        :param client: Optional. The client to use.  If not passed, falls back
+                       to the ``client`` stored on the ACL's parent.
         """
-        return self.save([])
+        self.save([], client=client)
+
+
+class BucketACL(ACL):
+    """An ACL specifically for a bucket.
+
+    :type bucket: :class:`gcloud.storage.bucket.Bucket`
+    :param bucket: The bucket to which this ACL relates.
+    """
+
+    def __init__(self, bucket):
+        super(BucketACL, self).__init__()
+        self.bucket = bucket
+
+    @property
+    def client(self):
+        """The client bound to this ACL's bucket."""
+        return self.bucket.client
+
+    @property
+    def reload_path(self):
+        """Compute the path for GET API requests for this ACL."""
+        return '%s/%s' % (self.bucket.path, self._URL_PATH_ELEM)
+
+    @property
+    def save_path(self):
+        """Compute the path for PATCH API requests for this ACL."""
+        return self.bucket.path
 
 
 class DefaultObjectACL(BucketACL):
     """A class representing the default object ACL for a bucket."""
 
     _URL_PATH_ELEM = 'defaultObjectAcl'
+    _PREDEFINED_QUERY_PARAM = 'predefinedDefaultObjectAcl'
 
 
 class ObjectACL(ACL):
-    """An ACL specifically for a key."""
+    """An ACL specifically for a Cloud Storage object / blob.
 
-    def __init__(self, key):
-        """
-        :type key: :class:`gcloud.storage.key.Key`
-        :param key: The key that this ACL corresponds to.
-        """
+    :type blob: :class:`gcloud.storage.blob.Blob`
+    :param blob: The blob that this ACL corresponds to.
+    """
+
+    def __init__(self, blob):
         super(ObjectACL, self).__init__()
-        self.key = key
+        self.blob = blob
 
-    def reload(self):
-        """Reload the ACL data from Cloud Storage.
+    @property
+    def client(self):
+        """The client bound to this ACL's blob."""
+        return self.blob.client
 
-        :rtype: :class:`ObjectACL`
-        :returns: The current ACL.
-        """
-        self.entities.clear()
+    @property
+    def reload_path(self):
+        """Compute the path for GET API requests for this ACL."""
+        return '%s/acl' % self.blob.path
 
-        url_path = '%s/acl' % self.key.path
-        found = self.key.connection.api_request(method='GET', path=url_path)
-        self.loaded = True
-        for entry in found['items']:
-            self.add_entity(self.entity_from_dict(entry))
-
-        return self
-
-    def save(self, acl=None):
-        """Save the ACL data for this key.
-
-        :type acl: :class:`gcloud.storage.acl.ACL`
-        :param acl: The ACL object to save.  If left blank, this will
-                    save the entries set locally on the ACL.
-        """
-        if acl is None:
-            acl = self
-            save_to_backend = acl.loaded
-        else:
-            save_to_backend = True
-
-        if save_to_backend:
-            result = self.key.connection.api_request(
-                method='PATCH', path=self.key.path, data={'acl': list(acl)},
-                query_params={'projection': 'full'})
-            self.entities.clear()
-            for entry in result['acl']:
-                self.add_entity(self.entity_from_dict(entry))
-            self.loaded = True
-
-        return self
-
-    def clear(self):
-        """Remove all ACL rules from the key.
-
-        Note that this won't actually remove *ALL* the rules, but it
-        will remove all the non-default rules.  In short, you'll still
-        have access to a key that you created even after you clear ACL
-        rules with this method.
-        """
-        return self.save([])
+    @property
+    def save_path(self):
+        """Compute the path for PATCH API requests for this ACL."""
+        return self.blob.path
