@@ -1,3 +1,17 @@
+# Copyright 2014 Google Inc. All rights reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 import unittest2
 
 
@@ -17,7 +31,7 @@ class Test_ACLEntity(unittest2.TestCase):
         self.assertEqual(entity.identifier, None)
         self.assertEqual(entity.get_roles(), set())
 
-    def test_ctor_explicit_identifier(self):
+    def test_ctor_w_identifier(self):
         TYPE = 'type'
         ID = 'id'
         entity = self._makeOne(TYPE, ID)
@@ -40,8 +54,7 @@ class Test_ACLEntity(unittest2.TestCase):
         TYPE = 'type'
         ROLE = 'role'
         entity = self._makeOne(TYPE)
-        found = entity.grant(ROLE)
-        self.assertTrue(found is entity)
+        entity.grant(ROLE)
         self.assertEqual(entity.get_roles(), set([ROLE]))
 
     def test_grant_duplicate(self):
@@ -58,8 +71,7 @@ class Test_ACLEntity(unittest2.TestCase):
         TYPE = 'type'
         ROLE = 'nonesuch'
         entity = self._makeOne(TYPE)
-        found = entity.revoke(ROLE)
-        self.assertTrue(found is entity)
+        entity.revoke(ROLE)
         self.assertEqual(entity.get_roles(), set())
 
     def test_revoke_hit(self):
@@ -136,6 +148,10 @@ class Test_ACL(unittest2.TestCase):
         acl._ensure_loaded()
         self.assertTrue(acl._really_loaded)
 
+    def test_client_is_abstract(self):
+        acl = self._makeOne()
+        self.assertRaises(NotImplementedError, lambda: acl.client)
+
     def test_reset(self):
         TYPE = 'type'
         ID = 'id'
@@ -370,7 +386,7 @@ class Test_ACL(unittest2.TestCase):
         from gcloud.storage.acl import _ACLEntity
         TYPE = 'type'
         ID = 'id'
-        KEY = '%s-%s' % (TYPE, ID)
+        ENTITY_VAL = '%s-%s' % (TYPE, ID)
         ROLE = 'role'
         entity = _ACLEntity(TYPE, ID)
         entity.grant(ROLE)
@@ -379,8 +395,8 @@ class Test_ACL(unittest2.TestCase):
         before = acl.entity(TYPE, ID)
         acl.add_entity(entity)
         self.assertTrue(acl.loaded)
-        self.assertFalse(acl.get_entity(KEY) is before)
-        self.assertTrue(acl.get_entity(KEY) is entity)
+        self.assertFalse(acl.get_entity(ENTITY_VAL) is before)
+        self.assertTrue(acl.get_entity(ENTITY_VAL) is entity)
         self.assertEqual(list(acl),
                          [{'entity': 'type-id', 'role': ROLE}])
         self.assertEqual(list(acl.get_entities()), [entity])
@@ -493,17 +509,183 @@ class Test_ACL(unittest2.TestCase):
         entity = acl.entity(TYPE, ID)
         self.assertEqual(acl.get_entities(), [entity])
 
-    def test_reload_raises_NotImplementedError(self):
+    def test_reload_missing(self):
+        # https://github.com/GoogleCloudPlatform/gcloud-python/issues/652
+        ROLE = 'role'
+        connection = _Connection({})
+        client = _Client(connection)
         acl = self._makeOne()
-        self.assertRaises(NotImplementedError, acl.reload)
+        acl.reload_path = '/testing/acl'
+        acl.loaded = True
+        acl.entity('allUsers', ROLE)
+        acl.reload(client=client)
+        self.assertEqual(list(acl), [])
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'GET')
+        self.assertEqual(kw[0]['path'], '/testing/acl')
 
-    def test_save_raises_NotImplementedError(self):
+    def test_reload_empty_result_clears_local(self):
+        ROLE = 'role'
+        connection = _Connection({'items': []})
+        client = _Client(connection)
         acl = self._makeOne()
-        self.assertRaises(NotImplementedError, acl.save)
+        acl.reload_path = '/testing/acl'
+        acl.loaded = True
+        acl.entity('allUsers', ROLE)
+        acl.reload(client=client)
+        self.assertTrue(acl.loaded)
+        self.assertEqual(list(acl), [])
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'GET')
+        self.assertEqual(kw[0]['path'], '/testing/acl')
+
+    def test_reload_nonempty_result(self):
+        ROLE = 'role'
+        connection = _Connection(
+            {'items': [{'entity': 'allUsers', 'role': ROLE}]})
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.reload_path = '/testing/acl'
+        acl.loaded = True
+        acl.reload(client=client)
+        self.assertTrue(acl.loaded)
+        self.assertEqual(list(acl), [{'entity': 'allUsers', 'role': ROLE}])
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'GET')
+        self.assertEqual(kw[0]['path'], '/testing/acl')
+
+    def test_save_none_set_none_passed(self):
+        connection = _Connection()
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.save(client=client)
+        kw = connection._requested
+        self.assertEqual(len(kw), 0)
+
+    def test_save_existing_missing_none_passed(self):
+        connection = _Connection({})
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.loaded = True
+        acl.save(client=client)
+        self.assertEqual(list(acl), [])
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'PATCH')
+        self.assertEqual(kw[0]['path'], '/testing')
+        self.assertEqual(kw[0]['data'], {'acl': []})
+        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
+
+    def test_save_no_acl(self):
+        ROLE = 'role'
+        AFTER = [{'entity': 'allUsers', 'role': ROLE}]
+        connection = _Connection({'acl': AFTER})
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.loaded = True
+        acl.entity('allUsers').grant(ROLE)
+        acl.save(client=client)
+        self.assertEqual(list(acl), AFTER)
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'PATCH')
+        self.assertEqual(kw[0]['path'], '/testing')
+        self.assertEqual(kw[0]['data'], {'acl': AFTER})
+        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
+
+    def test_save_w_acl(self):
+        ROLE1 = 'role1'
+        ROLE2 = 'role2'
+        STICKY = {'entity': 'allUsers', 'role': ROLE2}
+        new_acl = [{'entity': 'allUsers', 'role': ROLE1}]
+        connection = _Connection({'acl': [STICKY] + new_acl})
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.loaded = True
+        acl.save(new_acl, client=client)
+        entries = list(acl)
+        self.assertEqual(len(entries), 2)
+        self.assertTrue(STICKY in entries)
+        self.assertTrue(new_acl[0] in entries)
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'PATCH')
+        self.assertEqual(kw[0]['path'], '/testing')
+        self.assertEqual(kw[0]['data'], {'acl': new_acl})
+        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
+
+    def test_save_prefefined_invalid(self):
+        connection = _Connection()
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.loaded = True
+        with self.assertRaises(ValueError):
+            acl.save_predefined('bogus', client=client)
+
+    def test_save_predefined_valid(self):
+        PREDEFINED = 'private'
+        connection = _Connection({'acl': []})
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.loaded = True
+        acl.save_predefined(PREDEFINED, client=client)
+        entries = list(acl)
+        self.assertEqual(len(entries), 0)
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'PATCH')
+        self.assertEqual(kw[0]['path'], '/testing')
+        self.assertEqual(kw[0]['data'], {'acl': []})
+        self.assertEqual(kw[0]['query_params'],
+                         {'projection': 'full', 'predefinedAcl': PREDEFINED})
+
+    def test_save_predefined_valid_w_alternate_query_param(self):
+        # Cover case where subclass overrides _PREDEFINED_QUERY_PARAM
+        PREDEFINED = 'private'
+        connection = _Connection({'acl': []})
+        client = _Client(connection)
+        acl = self._makeOne()
+        acl.save_path = '/testing'
+        acl.loaded = True
+        acl._PREDEFINED_QUERY_PARAM = 'alternate'
+        acl.save_predefined(PREDEFINED, client=client)
+        entries = list(acl)
+        self.assertEqual(len(entries), 0)
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'PATCH')
+        self.assertEqual(kw[0]['path'], '/testing')
+        self.assertEqual(kw[0]['data'], {'acl': []})
+        self.assertEqual(kw[0]['query_params'],
+                         {'projection': 'full', 'alternate': PREDEFINED})
 
     def test_clear(self):
+        ROLE1 = 'role1'
+        ROLE2 = 'role2'
+        STICKY = {'entity': 'allUsers', 'role': ROLE2}
+        connection = _Connection({'acl': [STICKY]})
+        client = _Client(connection)
         acl = self._makeOne()
-        self.assertRaises(NotImplementedError, acl.clear)
+        acl.save_path = '/testing'
+        acl.loaded = True
+        acl.entity('allUsers', ROLE1)
+        acl.clear(client=client)
+        self.assertEqual(list(acl), [STICKY])
+        kw = connection._requested
+        self.assertEqual(len(kw), 1)
+        self.assertEqual(kw[0]['method'], 'PATCH')
+        self.assertEqual(kw[0]['path'], '/testing')
+        self.assertEqual(kw[0]['data'], {'acl': []})
+        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
 
 
 class Test_BucketACL(unittest2.TestCase):
@@ -516,123 +698,34 @@ class Test_BucketACL(unittest2.TestCase):
         return self._getTargetClass()(*args, **kw)
 
     def test_ctor(self):
-        bucket = object()
+        NAME = 'name'
+        bucket = _Bucket(NAME)
         acl = self._makeOne(bucket)
         self.assertEqual(acl.entities, {})
         self.assertFalse(acl.loaded)
         self.assertTrue(acl.bucket is bucket)
+        self.assertEqual(acl.reload_path, '/b/%s/acl' % NAME)
+        self.assertEqual(acl.save_path, '/b/%s' % NAME)
 
-    def test_reload_eager_empty(self):
-        NAME = 'name'
-        ROLE = 'role'
-        connection = _Connection({'items': []})
-        bucket = _Bucket(connection, NAME)
-        acl = self._makeOne(bucket)
-        acl.loaded = True
-        acl.entity('allUsers', ROLE)
-        self.assertTrue(acl.reload() is acl)
-        self.assertEqual(list(acl), [])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'GET')
-        self.assertEqual(kw[0]['path'], '/b/%s/acl' % NAME)
 
-    def test_reload_eager_nonempty(self):
-        NAME = 'name'
-        ROLE = 'role'
-        connection = _Connection(
-            {'items': [{'entity': 'allUsers', 'role': ROLE}]})
-        bucket = _Bucket(connection, NAME)
-        acl = self._makeOne(bucket)
-        acl.loaded = True
-        self.assertTrue(acl.reload() is acl)
-        self.assertEqual(list(acl), [{'entity': 'allUsers', 'role': ROLE}])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'GET')
-        self.assertEqual(kw[0]['path'], '/b/%s/acl' % NAME)
+class Test_DefaultObjectACL(unittest2.TestCase):
 
-    def test_reload_lazy(self):
-        NAME = 'name'
-        ROLE = 'role'
-        connection = _Connection(
-            {'items': [{'entity': 'allUsers', 'role': ROLE}]})
-        bucket = _Bucket(connection, NAME)
-        acl = self._makeOne(bucket)
-        self.assertTrue(acl.reload() is acl)
-        self.assertEqual(list(acl), [{'entity': 'allUsers', 'role': ROLE}])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'GET')
-        self.assertEqual(kw[0]['path'], '/b/%s/acl' % NAME)
+    def _getTargetClass(self):
+        from gcloud.storage.acl import DefaultObjectACL
+        return DefaultObjectACL
 
-    def test_save_none_set_none_passed(self):
-        NAME = 'name'
-        connection = _Connection()
-        bucket = _Bucket(connection, NAME)
-        acl = self._makeOne(bucket)
-        self.assertTrue(acl.save() is acl)
-        kw = connection._requested
-        self.assertEqual(len(kw), 0)
+    def _makeOne(self, *args, **kw):
+        return self._getTargetClass()(*args, **kw)
 
-    def test_save_no_arg(self):
+    def test_ctor(self):
         NAME = 'name'
-        ROLE = 'role'
-        AFTER = [{'entity': 'allUsers', 'role': ROLE}]
-        connection = _Connection({'acl': AFTER})
-        bucket = _Bucket(connection, NAME)
+        bucket = _Bucket(NAME)
         acl = self._makeOne(bucket)
-        acl.loaded = True
-        acl.entity('allUsers').grant(ROLE)
-        self.assertTrue(acl.save() is acl)
-        self.assertEqual(list(acl), AFTER)
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'PATCH')
-        self.assertEqual(kw[0]['path'], '/b/%s' % NAME)
-        self.assertEqual(kw[0]['data'], {'acl': AFTER})
-        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
-
-    def test_save_w_arg(self):
-        NAME = 'name'
-        ROLE1 = 'role1'
-        ROLE2 = 'role2'
-        STICKY = {'entity': 'allUsers', 'role': ROLE2}
-        new_acl = [{'entity': 'allUsers', 'role': ROLE1}]
-        connection = _Connection({'acl': [STICKY] + new_acl})
-        bucket = _Bucket(connection, NAME)
-        acl = self._makeOne(bucket)
-        acl.loaded = True
-        self.assertTrue(acl.save(new_acl) is acl)
-        entries = list(acl)
-        self.assertEqual(len(entries), 2)
-        self.assertTrue(STICKY in entries)
-        self.assertTrue(new_acl[0] in entries)
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'PATCH')
-        self.assertEqual(kw[0]['path'], '/b/%s' % NAME)
-        self.assertEqual(kw[0]['data'], {'acl': new_acl})
-        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
-
-    def test_clear(self):
-        NAME = 'name'
-        ROLE1 = 'role1'
-        ROLE2 = 'role2'
-        STICKY = {'entity': 'allUsers', 'role': ROLE2}
-        connection = _Connection({'acl': [STICKY]})
-        bucket = _Bucket(connection, NAME)
-        acl = self._makeOne(bucket)
-        acl.loaded = True
-        acl.entity('allUsers', ROLE1)
-        self.assertTrue(acl.clear() is acl)
-        self.assertEqual(list(acl), [STICKY])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'PATCH')
-        self.assertEqual(kw[0]['path'], '/b/%s' % NAME)
-        self.assertEqual(kw[0]['data'], {'acl': []})
-        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
+        self.assertEqual(acl.entities, {})
+        self.assertFalse(acl.loaded)
+        self.assertTrue(acl.bucket is bucket)
+        self.assertEqual(acl.reload_path, '/b/%s/defaultObjectAcl' % NAME)
+        self.assertEqual(acl.save_path, '/b/%s' % NAME)
 
 
 class Test_ObjectACL(unittest2.TestCase):
@@ -645,146 +738,32 @@ class Test_ObjectACL(unittest2.TestCase):
         return self._getTargetClass()(*args, **kw)
 
     def test_ctor(self):
-        key = object()
-        acl = self._makeOne(key)
+        NAME = 'name'
+        BLOB_NAME = 'blob-name'
+        bucket = _Bucket(NAME)
+        blob = _Blob(bucket, BLOB_NAME)
+        acl = self._makeOne(blob)
         self.assertEqual(acl.entities, {})
         self.assertFalse(acl.loaded)
-        self.assertTrue(acl.key is key)
-
-    def test_reload_eager_empty(self):
-        NAME = 'name'
-        KEY = 'key'
-        ROLE = 'role'
-        after = {'items': [{'entity': 'allUsers', 'role': ROLE}]}
-        connection = _Connection(after)
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        acl.loaded = True
-        self.assertTrue(acl.reload() is acl)
-        self.assertEqual(list(acl), after['items'])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'GET')
-        self.assertEqual(kw[0]['path'], '/b/name/o/%s/acl' % KEY)
-
-    def test_reload_eager_nonempty(self):
-        NAME = 'name'
-        KEY = 'key'
-        ROLE = 'role'
-        after = {'items': []}
-        connection = _Connection(after)
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        acl.loaded = True
-        acl.entity('allUsers', ROLE)
-        self.assertTrue(acl.reload() is acl)
-        self.assertEqual(list(acl), [])
-
-    def test_reload_lazy(self):
-        NAME = 'name'
-        KEY = 'key'
-        ROLE = 'role'
-        after = {'items': [{'entity': 'allUsers', 'role': ROLE}]}
-        connection = _Connection(after)
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        self.assertTrue(acl.reload() is acl)
-        self.assertEqual(list(acl),
-                         [{'entity': 'allUsers', 'role': ROLE}])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'GET')
-        self.assertEqual(kw[0]['path'], '/b/name/o/%s/acl' % KEY)
-
-    def test_save_none_set_none_passed(self):
-        NAME = 'name'
-        KEY = 'key'
-        connection = _Connection()
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        self.assertTrue(acl.save() is acl)
-        kw = connection._requested
-        self.assertEqual(len(kw), 0)
-
-    def test_save_existing_set_none_passed(self):
-        NAME = 'name'
-        KEY = 'key'
-        connection = _Connection({'foo': 'Foo', 'acl': []})
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        acl.loaded = True
-        self.assertTrue(acl.save() is acl)
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'PATCH')
-        self.assertEqual(kw[0]['path'], '/b/name/o/%s' % KEY)
-        self.assertEqual(kw[0]['data'], {'acl': []})
-        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
-
-    def test_save_existing_set_new_passed(self):
-        NAME = 'name'
-        KEY = 'key'
-        ROLE = 'role'
-        new_acl = [{'entity': 'allUsers', 'role': ROLE}]
-        connection = _Connection({'foo': 'Foo', 'acl': new_acl})
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        acl.loaded = True
-        acl.entity('allUsers', 'other-role')
-        self.assertTrue(acl.save(new_acl) is acl)
-        self.assertEqual(list(acl), new_acl)
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'PATCH')
-        self.assertEqual(kw[0]['path'], '/b/name/o/%s' % KEY)
-        self.assertEqual(kw[0]['data'], {'acl': new_acl})
-        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
-
-    def test_clear(self):
-        NAME = 'name'
-        KEY = 'key'
-        ROLE = 'role'
-        connection = _Connection({'foo': 'Foo', 'acl': []})
-        bucket = _Bucket(connection, NAME)
-        key = _Key(bucket, KEY)
-        acl = self._makeOne(key)
-        acl.loaded = True
-        acl.entity('allUsers', ROLE)
-        self.assertTrue(acl.clear() is acl)
-        self.assertEqual(list(acl), [])
-        kw = connection._requested
-        self.assertEqual(len(kw), 1)
-        self.assertEqual(kw[0]['method'], 'PATCH')
-        self.assertEqual(kw[0]['path'], '/b/name/o/%s' % KEY)
-        self.assertEqual(kw[0]['data'], {'acl': []})
-        self.assertEqual(kw[0]['query_params'], {'projection': 'full'})
+        self.assertTrue(acl.blob is blob)
+        self.assertEqual(acl.reload_path, '/b/%s/o/%s/acl' % (NAME, BLOB_NAME))
+        self.assertEqual(acl.save_path, '/b/%s/o/%s' % (NAME, BLOB_NAME))
 
 
-class _Key(object):
+class _Blob(object):
 
-    def __init__(self, bucket, key):
+    def __init__(self, bucket, blob):
         self.bucket = bucket
-        self.key = key
-
-    @property
-    def connection(self):
-        return self.bucket.connection
+        self.blob = blob
 
     @property
     def path(self):
-        return '%s/o/%s' % (self.bucket.path, self.key)
+        return '%s/o/%s' % (self.bucket.path, self.blob)
 
 
 class _Bucket(object):
 
-    def __init__(self, connection, name):
-        self.connection = connection
+    def __init__(self, name):
         self.name = name
 
     @property
@@ -801,7 +780,7 @@ class _Connection(object):
         self._deleted = []
 
     def api_request(self, **kw):
-        from gcloud.storage.exceptions import NotFound
+        from gcloud.exceptions import NotFound
         self._requested.append(kw)
 
         try:
@@ -810,3 +789,9 @@ class _Connection(object):
             raise NotFound('miss')
         else:
             return response
+
+
+class _Client(object):
+
+    def __init__(self, connection):
+        self.connection = connection
